/****************************************************************************
*****************************************************************************
**   The Array Extender                                                    **
*****************************************************************************
*****************************************************************************
**                                                                         **
**  By: Alan Kreutzer  --                                                  **
**  May 2002                                                               **
**                                                                         **
**  Requires WinBatch version 2001 or later                                **
**                                                                         **
**  May be freely distrubuted.                                             **
**  This program is distributed with no warrenty, use at your own risk.    **
**                                                                         **
*****************************************************************************
****************************************************************************/
//#define EXTENDERVERSION           // Read Version from RC file,  #1001

#define INTERFACELEVEL  2           // extender interface level 
#define EXTENDER_FEATURE_LEVEL 1001 // extender feature level 

#define STRICT


/*---------------------------------------------------------------------------+
!                                   Includes                                 !
+---------------------------------------------------------------------------*/
#include <windows.h>
                           /* special includes for WinBatch extenders */
#include "wwwpform.h"
#include "addons.h"


/*---------------------------------------------------------------------------+
!       Definitions used when manipulating arrays.                           !
+---------------------------------------------------------------------------*/
#define VARTYPE_UNDEF      0           // These are bitmapped, var can be both
#define VARTYPE_INT        1           // Numeric and string, must check bits
#define VARTYPE_STRING     2
#define VARTYPE_FILE       5           // FILE and INT bits set
#define VARTYPE_UNINIT     8
#define VARTYPE_OLEOBJECT 17           // OLE AND INT bits set
#define VARTYPE_FLOATNUM  32           // floating point
#define VARTYPE_BINARY    65           // BINARY and INT bits set
#define VARTYPE_XOLEOBJECT 129         /// an OLE object that has been temporarily camouflaged
#define VARTYPE_ARRAY     256          /// an array

typedef struct arrayvariable
{
  double Aflt;                         // 8-byte floating point number
  LONG   Ax;                           // if int variable, value is here
  int    Ainittype;                    // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  DWORD  Astrptr;                      // pointer to string space for string
} arrvar;


typedef struct resultants2
{
  union u
  {
    struct s                           // for normal variables
    {
      double flt;                      // 8-byte floating point number
      LONG   x;                        // if int variable, value is here
    };
    struct ss                          // for VARTYPE_ARRAY
    {
      WORD wDim[5];                    // array dimension sizes
      BYTE cDims;                      // number of array dimensions
      BYTE cFlags;                     // array flags
    };
  };
  int   inittype;                      // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  union uu
  {
    DWORD strptr;                      // pointer to string space for string
    LPSTR lpBuffer;                    // pointer to locked array data buffer for VARTYPE_ARRAY
  };
  WORD  vorigin;                       // index to var area is derived from var, or FFFF
} resultvar2;

typedef arrvar FAR *LPARRVAR;

typedef resultvar2 FAR * LPRESULTVAR2;
   
/*---------------------------------------------------------------------------+
!       Common data block, passed to called routines                         !  
+---------------------------------------------------------------------------*/
typedef struct xCommonStruct
{
     long        Long;       /* Numeric value (when return = 'L'            */
     double      Float;      /* Floating point(when return = 'F'            */
     char        *String;    /* String        (when return = 'S' or 'E')    */
     resultvar2  Array;      /* Array         (when return = 'A')           */
     char        *Buffer;    /* Memory Buffer                               */
     long        BuffSize;   /* Size of Buffer                              */
     BOOL        Borrowed;   /* Is buffer borrowed memory?                  */
     long        ErrorType;  /* Error type... -1 (Normal) or -3 (Fatal)     */
     char        Type;       /* A = Array, S = String, L = Long, E = Error. */
} xCommon;

/*---------------------------------------------------------------------------+
!                           Function Prototypes                              !
+---------------------------------------------------------------------------*/

BOOL  GetBuffer(xCommon *Common, long Size);
long  CompareStrings(char *A, char *B);
long  CompareStringsNC(char *A, char *B);
LPSTR GetArrayString(xCommon *Common, LPARRVAR ArrayItem); 
BOOL  PutArrayString(xCommon *Common, LPARRVAR lpArray, char *TheString);
long  ArrayItemType(LPARRVAR ArrayItem);
LPARRVAR ArrayElement(LPRESULTVAR2 arr, WORD x0, WORD x1, WORD x2, WORD x3, WORD x4);
void  ArraySorter(xCommon *Common, LPRESULTVAR2 ArrIn, long (*CompString)(char *, char *));

/*---------------------------------------------------------------------------+
!                Debug Functions                                             !
+---------------------------------------------------------------------------*/
void FAR PASCAL MBSS(char *Title, char *Text);  
void FAR PASCAL MBSN(char *Title, long Number);
void FAR PASCAL MBSF(char *Title, double Flt);
void FAR PASCAL MBSX(char *Title, long Number);



/*---------------------------------------------------------------------------+
!                           Global Variables                                 !
+---------------------------------------------------------------------------*/
         /*--------- Superstitious data field to bypass/ignore bugs in
         !    certain versions of Windows that cause the beginning 
         !    of the data area to be overwritten.  This field should 
         !    be the first data area defined and linked. 
         +------------------*/
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE   hinstDLL;                    // DLL Instance saved
long        ExtenderVersion;
DLLSERVICES lpfnDllServices = NULL;      // Helper routine from WinBatch
HGLOBAL     hBatData = NULL;             // Passed from WinBatch

        /*------------------------------------------------------+
        !       Case conversion table.                          !
        !       (Converts upper to lower case)                  !
        +------------------------------------------------------*/
#define LOWERCASE(C) LowerTable[(unsigned char) C]
unsigned char LowerTable[] =
    {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
     0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
     0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
     0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,
     0x40,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x5b,0x5c,0x5d,0x5e,0x5f,
     0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x7b,0x7c,0x7d,0x7e,0x7f,
     0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x9a,0x8b,0x9c,0x8d,0x8e,0x8f, 
     0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0xFf, 
     0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAa,0xAb,0xAc,0xAd,0xAe,0xAf,
     0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBa,0xBb,0xBc,0xBd,0xBe,0xBf, 
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xD7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xDf,
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xFf}; 





/***************************************************************************
*                                                                          *
*                                                                          *
*                  Command Extension Table                                 *
*                                                                          *
*                                                                          *
***************************************************************************/
tablestruct CommandExtTable[] =
{
//----------------------------------------------------------------------
// ID   Type       Mask              Params/Value, Size, Name
//----------------------------------------------------------------------
  {10, FUNCTION, 0x0000000000000000, 0x00000001L, 10, "ArrayExtenderVersion" }, 
  {11, FUNCTION, 0x0000000000000003, 0x00000001L,  9, "ArraySort"            },
  {12, FUNCTION, 0x0000000000000003, 0x00000001L, 11, "ArraySortNC"          },
  {13, FUNCTION, 0x0000000000000003, 0x00000001L, 14, "ArrayDuplicate"       },
  {14, FUNCTION, 0x00000000000000F3, 0x00000403L, 11, "ArrayResize"          }, 
  {15, FUNCTION, 0x0000000000000001, 0x00000001L, 13, "ArrayizeBytes"        },
//--------------------------------------------------------------------------
  {-1,        0,                  0, 0x00000000,   0, ""                }
};




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                               DllMain                                      **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);
  return (TRUE);
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                        WILExtenderQuery                                    **
**                                                                            **
**           Tells the WilDll about our interface level.                      **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*----------------------------------------------------------------------+
!   case 1 ... identify yourself to a caller.                           !
+----------------------------------------------------------------------*/
    case 1:                      
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }
      return ExtenderVersion;
    }

/*----------------------------------------------------------------------+
!   case 2 ... Tell WinBatch we use interface level 2                   !   
+----------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }


/*----------------------------------------------------------------------+
!    unknown requests                                           !  
+----------------------------------------------------------------------*/
    default:
      return (0);
  }
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                          WILExtender2                                      **
**                                                                            **
**                                                                            **
**                                                                            **
**   The main routine performs all extender functions, and various            **
**   initialization and cleanup tasks.                                        **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/

DllExport LONG FAR PASCAL WILExtender2
  (HWND           hWnd,                   // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{  
  xCommon      Common;
  LPRESULTVAR2 ArrIn;
  char         szWork[256];   
  long         J, K;
  char         *Ix, *Ox;
  char         C;

  Common.Array.lpBuffer = NULL;
  Common.Buffer = szWork;
  Common.BuffSize = 256;
  Common.Borrowed = FALSE;
  Common.ErrorType = -1;
  
    
     /***************************************************
     *                                                  *
     *    The "Big Switch."                             *    
     *                                                  *
     *    Calls from Winbatch include an integer value  *
     *    (CmdId) which selects a function from this    *
     *    switch.   Calls < 0 are WinBatch overhead.    *
     *    Calls > 0 are our functions.                  *
     *                                                  *
     ***************************************************/             
  switch (CmdId)         
  {
  
/*----------------------------------------------------------------------+
!  case -1 ... One time initialization routine.                         !
!                                                                       !
!              Return the number of functions  & constants defined.     !
+----------------------------------------------------------------------*/
    case -1:                         
    {
       lpfnDllServices = lpViperStruct->lpfnDllServices;
       hBatData = lpViperStruct->hBatData;
       LoadString(hinstDLL, 1001, szWork, sizeof(szWork));
       Ix = szWork;
       ExtenderVersion = 0;
       while ((C = *Ix++) != '\0') ExtenderVersion = (ExtenderVersion * 10) + (C - '0'); 
       return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1); 
    }

/*----------------------------------------------------------------------+
!  case -2 ... Called repeatedly as extender is being loaded.           !
!                                                                       !
!              On each call, return a function table entry.             !
+----------------------------------------------------------------------*/
    case -2:                          
    {
      LPTABLESTRUCT lpTable;

      lpTable = (LPTABLESTRUCT) lpvvArg;
      J = (UINT) nArgCount;

      if (CommandExtTable[J].ident == -1)
        return (0);

      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[J].ident;
        lpTable->type   = CommandExtTable[J].type;
        lpTable->param1 = CommandExtTable[J].param1;
        lpTable->ulMask = CommandExtTable[J].ulMask;
        K = 0;
        Ix = (char *) CommandExtTable[J].delim;
        Ox = (char *) lpTable->delim;
        while ((*Ox++ = LOWERCASE(*Ix++)) != '\0') K++;
        lpTable->size = K;
      }
      return (1);
    }


/*----------------------------------------------------------------------+
!  case -3 .. Called once at shutdown. Can be used for cleanup.         !
+----------------------------------------------------------------------*/
    case -3:
    {
        return (1);
    }


/*----------------------------------------------------------------------+
!  case 10 ... ArrxVersion                                              !
!                                                                       !
!             Return extender version information.                      !
!                                                                       !
!              Parameter 0 = Version                                    !
!                        1 = Total Functions                            !
!                        2 = Total Constants                            !
+----------------------------------------------------------------------*/
    case 10:
    {       
         UINT uType;
         
         if (lpvvArg[0].x == 0)             // Check the request number
         {
              Common.Type = 'L';
              Common.Long = ExtenderVersion;
              break;
         } 
   
         if (lpvvArg[0].x == 1)            // total functions
             uType = FUNCTION;
         else 
            if (lpvvArg[0].x == 2)         // total constants
                uType = CONSTANT;
            else
            {
                Common.Type = 'E';          // Invalid parameter
                Common.Long = 101;
                break;
            }

        Common.Type = 'L';
        Common.Long = 0;
        for (K = 0; CommandExtTable[K].ident != -1; K++)
           if ((UINT) CommandExtTable[K].type == uType)
                Common.Long++;
        break;
    }

/*----------------------------------------------------------------------+
!  case 11 ... ArraySort(Array)                                         !
!                                                                       !
!         Sort an array on column 1.    (Case Sensitive)                !
+----------------------------------------------------------------------*/ 
    case 11:
    {      
        Common.Type = 'L';
        Common.Long = 1;    
        ArraySorter(&Common, (LPRESULTVAR2) &lpvvArg[0], &CompareStrings);
        break;
    }

/*----------------------------------------------------------------------+
!  case 12 ... ArraySortNC(Array)                                       !
!                                                                       !
!         Sort an array on column 1.    (Not case sensitive)            !
+----------------------------------------------------------------------*/ 
    case 12:
    {      
        Common.Type = 'L';
        Common.Long = 1;    
        ArraySorter(&Common, (LPRESULTVAR2) &lpvvArg[0], &CompareStringsNC);
        break;
    }
    
/*----------------------------------------------------------------------+
!  case 13 ... ArrayDuplicate(Array)                                    !
!                                                                       !
!       Returns a duplicate of the array.                               !
!                                                                       !
!       In WinBatch, B = ARRAY   doesn't create a duplicte of an array. !
!       It creates a pointer, so that A and B both point at the same    !
!       array, and any changes to one also changes the other.           !
!                                                                       !
!       ArrayDuplicate returns a true duplicate so you can change the   !
!       new array without changing the old one.                         !
+----------------------------------------------------------------------*/ 
    case 13:
    {      
       LPARRVAR From;
       LPARRVAR To;
       long     ItemCount;
       
       Common.Type = 'A';
       
       ArrIn = (LPRESULTVAR2) &lpvvArg[0];
      
       Common.Array.cDims      = ArrIn->cDims;
       Common.Array.wDim[0]    = ArrIn->wDim[0];
       Common.Array.wDim[1]    = ArrIn->wDim[1];
       Common.Array.wDim[2]    = ArrIn->wDim[2];
       Common.Array.wDim[3]    = ArrIn->wDim[3];
       Common.Array.wDim[4]    = ArrIn->wDim[4];
       Common.Array.cFlags     = ArrIn->cFlags;
       
       ItemCount =  Common.Array.wDim[0];    
       for (J=1; J<Common.Array.cDims;J++) ItemCount = ItemCount * Common.Array.wDim[J];
       
       Common.Array.lpBuffer  =  (LPSTR) GlobalAlloc(GPTR, ItemCount * sizeof(arrvar));
       if (Common.Array.lpBuffer == NULL)
       {
           Common.Type = 'E';
           Common.Long = 913;
           Common.ErrorType = -3;
           break;
       }

       From = (LPARRVAR) ArrIn->lpBuffer;
       To   = (LPARRVAR) Common.Array.lpBuffer;
       CopyMemory(To, From,ItemCount * sizeof(arrvar)); 
       
       for (J=0;J<ItemCount;J++)
       {
           if (To->Ainittype & VARTYPE_STRING)
           {
               Ix = GetArrayString(&Common, From);
               if (Ix != NULL) PutArrayString(&Common, To, Ix);
               if (Common.Type == 'E') break;
           }
           From  += 1;
           To    += 1;
       }
       break;
    }    
    
/*----------------------------------------------------------------------+
!  case 14 ... ArrayResize(Array,InitialValue,1,[2,3,4,5])              !
!                                                                       !
!         Returns an array of the specified dimensions.                 !
!         The new array is initialized with the contents of the         !
!         passed array.  If the new array is smaller than the old       !
!         array, elements will be lost.  If it is larger, the new       !
!         elements are initialized with the passed "InitialValue"       !
!                                                                       !
!         The passed array is not modified.                             !
+----------------------------------------------------------------------*/ 
    case 14:
    {      
       LPARRVAR From;
       LPARRVAR To;
       long     ItemCount;
       long     MaxDim;
       long     x[5];
       BOOLEAN  Done;
       
       int   InitialType;
       LONG  InitialValue;
       
       Common.Type = 'A';
       
       //----------------------------------------------------------------
       // Get the address of the old array, and the size of the new one.
       //---------------------------------------------------------------- 
       ArrIn = (LPRESULTVAR2) &lpvvArg[0];
       for (J=0;J<5;J++)         Common.Array.wDim[J] = 0;
       for (J=0;J<(nArgCount-2);J++) Common.Array.wDim[J] = lpvvArg[J+2].x;

       Common.Array.cDims = 0;        
       for (J=0;J<5;J++)
       {
           if (Common.Array.wDim[J] == 0) break;
           Common.Array.cDims++;
       }
       
       //----------------------------------------------------------------
       //  Compute the number of items in the new array.
       //---------------------------------------------------------------- 
       ItemCount = Common.Array.wDim[0];    
       for (J=1; J<Common.Array.cDims;J++) 
          ItemCount = ItemCount * Common.Array.wDim[J];

       if (ItemCount < 1)
       {
          Common.Type = 'E';
          Common.Long = 102;
          break;
       }
       
       //----------------------------------------------------------------
       // Get memory for the new array.
       //---------------------------------------------------------------- 
       Common.Array.lpBuffer  =  (LPSTR) GlobalAlloc(GPTR, ItemCount * sizeof(arrvar));
       if (Common.Array.lpBuffer == NULL)
       {
           Common.Type = 'E';
           Common.Long = 913;
           Common.ErrorType = -3;
           break;
       }
       
       //----------------------------------------------------------------
       // initialize everything in the new array with an invalid value
       // so we can tell, later, which ones we didn't copy from the
       // old array.
       //---------------------------------------------------------------- 
       To = (LPARRVAR) Common.Array.lpBuffer;
       for (J=0;J<ItemCount;J++)
       {
           To->Ainittype = -1;
           To++;
       }

       //---------------------------------------------------------------- 
       //  Copy corresponding items from the old array to the new array.
       //---------------------------------------------------------------- 
       MaxDim = ((ArrIn->cDims<Common.Array.cDims)?ArrIn->cDims:Common.Array.cDims)-1;
       x[0] = x[1] = x[2] = x[3] = x[4] = 0;
       Done = FALSE;
       while(!Done)
       {
           From = ArrayElement(ArrIn,x[0],x[1],x[2],x[3],x[4]);
           if (From != NULL)
           {
              To = ArrayElement(&Common.Array,x[0],x[1],x[2],x[3],x[4]);
              if (To != NULL)
              {
                  CopyMemory(To,From,sizeof(arrvar)); 
                  if (To->Ainittype & VARTYPE_STRING)
                  {
                    Ix = GetArrayString(&Common, From);
                    if (Ix != NULL) PutArrayString(&Common, To, Ix);
                  }
              }
           }
           if (Common.Type == 'E') break;
           J=MaxDim;
           while (1)
           {
              x[J]++;
              if (x[J] < Common.Array.wDim[J]) break;
              if (J == 0){Done=TRUE; break;}
              x[J] = 0;
              J--;
           }
       } 

       //---------------------------------------------------------------- 
       //  Find all the items we didn't copy and set them to the  
       //  user's "initial value"
       //----------------------------------------------------------------  
       InitialType = lpvvArg[1].inittype;
       
       To = (LPARRVAR) Common.Array.lpBuffer;
       if (InitialType & VARTYPE_STRING)      
       {
           for (J=0;J<ItemCount;J++)
           {
               if (To->Ainittype == -1) 
                  PutArrayString(&Common,To,lpvvArg[1].lpstr);
               To++;
           }
           break;
       }
       
       if (InitialType & VARTYPE_FLOATNUM)
       {
           for (J=0;J<ItemCount;J++)
           {
               if (To->Ainittype == -1)
               {
                   To->Ainittype = VARTYPE_FLOATNUM;
                   To->Aflt = lpvvArg[1].flt;
               }
               To++;
           }
           break;
       }
       
       InitialValue = lpvvArg[1].x;
       if (!(InitialType & VARTYPE_INT)) 
       {
          InitialType = VARTYPE_INT;
          InitialValue = 0;
       }
       for (J=0;J<ItemCount;J++)
       {
           if (To->Ainittype == -1)
           {
               To->Ax = InitialValue;
               To->Ainittype = InitialType;
           }
           To++;
       }      
       break;                   
    }
   
    
/*----------------------------------------------------------------------+
!  case 15 ... ArrayizeBytes(String)                                    !
!                                                                       !
!       Returns an array, each member containing one byte from          !
!       the string.                                                     !
!                                                                       !
+----------------------------------------------------------------------*/
    case 15:
    {      
       LPARRVAR Next;
       char     OneByte[2];

       Common.Type = 'A';
       
       
       Ix = lpvvArg[0].lpstr;
       if (*Ix == '\0') Ix = " ";
       K = lstrlen(Ix);
      
       Common.Array.cDims      = 1;
       Common.Array.wDim[0]    = K;
       Common.Array.wDim[1]    = 0;
       Common.Array.wDim[2]    = 0;
       Common.Array.wDim[3]    = 0;
       Common.Array.wDim[4]    = 0;
       Common.Array.cFlags     = 0;
       
       Common.Array.lpBuffer  = (LPSTR) GlobalAlloc(GPTR, K * sizeof(arrvar));
       if (Common.Array.lpBuffer == NULL)
       {
           Common.Type = 'E';
           Common.Long = 913;
           Common.ErrorType = -3;
           break;
       }
       
       OneByte[1] = '\0';
       Next = (LPARRVAR) Common.Array.lpBuffer;
       for (J=0;J<K;J++)
       {
           OneByte[0] = *Ix++;
           PutArrayString(&Common,Next,OneByte);
           Next++; 
       }
       break;
    } 
           
/*----------------------------------------------------------------------+
!   Default Case.  (We should never get here)                           !
+-----------------------------------------------------------------------*/
    default:      
    {
        Common.Type = 'E';
        Common.Long = 100;
        break;
    }
  }                   
  
  
            /*************************************
            *                                    *
            *       End of "Big Switch"          *
            *                                    *
            **************************************
 
 
 
 
 
/*-----------------------------------------------------------------------+
!+----------------------------------------------------------------------+!
!!                                                                      !!
!!   Return the value,  free allocated memory, and return.              !!
!!                                                                      !!
!+----------------------------------------------------------------------+!
+-----------------------------------------------------------------------*/ 
 switch (Common.Type)         
  { 
      case 'A':   // Return an array
    {
       lpViperReturn->inittype = VARTYPE_ARRAY;
       lpViperReturn->flt      = Common.Array.flt;
       lpViperReturn->x        = Common.Array.x;
       lpViperReturn->lpstr    = Common.Array.lpBuffer;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
  
    case 'L':  // Return a long integer
    {
       lpViperReturn->inittype=VARTYPE_INT;
       lpViperReturn->x=Common.Long;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
    
    case 'S':   // Return a string
    {
       lpViperReturn->inittype=VARTYPE_STRING;
       lpViperReturn->lpstr=Common.String;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);    
       break;
    }

    case 'F':   // Return a floating point number
    {
       lpViperReturn->inittype=VARTYPE_FLOATNUM;
       lpViperReturn->flt=Common.Float;
       J = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
        
    case 'E':  // Return an error message
    {
       LoadString(hinstDLL, Common.Long, szWork, sizeof(szWork));
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = szWork;
       J = lpViperCallback(Common.ErrorType, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
    
    default:  // Invalid return type.  (This better not happen!)
    {
       LoadString(hinstDLL, 900, szWork, sizeof(szWork));
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = szWork;
       J = lpViperCallback(-3, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }  
  }

  // Free borrowed memory
    
  if (Common.Borrowed) GlobalFree(Common.Buffer); 
  if (Common.Array.lpBuffer != NULL) GlobalFree(Common.Array.lpBuffer);
  return J;
}
  /************************************
   *                                  *
   *   End of program WILExtender2    *
   *                                  *
   ***********************************/








/************************************************************************
*                                                                       *
*  *******************************************************************  *
*  *                                                                 *  *
*  *                                                                 *  *
*  *          Helper routines.                                       *  *
*  *                                                                 *  *
*  *                                                                 *  *
*  *******************************************************************  *
*                                                                       *
*************************************************************************/         






/****************************************************************************
*  Get a buffer.                                                            *
*                                                                           *
*  If the default buffer is long enough, return.                            *
*  Otherwise allocate a bigger one.                                         *
*                                                                           *
*  The Current buffer address is at  Common->Buffer                         *
*  The size of the buffer is at      Common->BuffSize                       *
*                                                                           *
*  Common->Borrowed is FALSE for the default buffer, but is set TRUE        *
*  when we allocate a new buffer.  (This flag will remind us that we need   *
*  to FREE the buffer at exit time.)                                        * 
*                                                                           *
*  Note... We are passed the size of the longest string the buffer will     *
*  need to hold.  We must allocate size+1 bytes,  so there will be room     *
*  for the string and its terminating \0 character.                         *
****************************************************************************/
BOOL GetBuffer(xCommon *Common, long Size)
{
                         
      if (Size < Common->BuffSize)           // If current buffer is big enough        
           return (TRUE);                    // return with no change.

      if (Common->Borrowed)                  // Do we have a borrowed buffer?
      {                                      // If so, free it before we
           GlobalFree(Common->Buffer);       // borrow a new one.
           Common->Borrowed = FALSE;
      }
           
      Size = Size + 1;    
      Common->Buffer = GlobalAlloc(GMEM_FIXED,Size);   // Borrow memory.
      if (Common->Buffer == NULL) 
      {
           Common->Type = 'E';
           Common->Long = 913;
           Common->ErrorType = -3;
           return FALSE;
      }
      Common->Borrowed = TRUE;
      Common->BuffSize = Size;
      return TRUE;
}

	
/****************************************************************************
*                                                                           *
*    CompareStrings(stringA, stringB)                                       *
*                                                                           *
*     Does a case-sensitive compare of two strings.                         *
*     (Trailing spaces are not considered - strings that are                *
*      equal except for trailing spaces are equal.)                         *
*                                                                           *
*     Returns  1 if string A is greater than B                              *
*              0 if the strings are equal                                   *
*             -1 if string A is less than B                                 *
*                                                                           *
****************************************************************************/
long CompareStrings(char *A, char *B)
{
	unsigned char X;
	unsigned char Y;
    
     //  Walk through the strings, comparing characters
     //  (with all lower case converted to upper case)

	while (1) {
		X = (unsigned char) *A++;
		Y = (unsigned char) *B++;
		if (X == '\0') break;
		if (Y == '\0') break;
		if (X > Y) return 1;
		if (X < Y) return -1;
	}
    //  If string 1 ended... search the remainder of string 2.
	//  Anything other than a space here means string 1 is smaller.

	if (X == '\0') {
        while(1)  {
			if (Y == '\0') return 0;
			if (Y != ' ') return -1;
		    Y = (unsigned char) *B++;
        }
    }
    
     //  If string 2 ended, search the remainder of string 1.
     // Anything other than a space means string 1 is larger.

	while (1) {
		if(X == '\0') return 0;
		if(X != ' ') return 1;
		X =(unsigned char) *A++;
	}		
}

	
/****************************************************************************
*                                                                           *
*    CompareStringsNC(stringA, stringB)                                     *
*                                                                           *
*     Does a case-insensitive compare of two strings.                       *
*     (Trailing spaces are not considered - strings that are                *
*      equal except for trailing spaces are equal.)                         *
*                                                                           *
*     Returns  1 if string A is greater than B                              *
*              0 if the strings are equal                                   *
*             -1 if string A is less than B                                 *
*                                                                           *
****************************************************************************/
long CompareStringsNC(char *A, char *B)
{
	unsigned char X;
	unsigned char Y;
    
     //  Walk through the strings, comparing characters
     //  (with all lower case converted to upper case)

	while (1) {
		X = (unsigned char) *A++;
		Y = (unsigned char) *B++;
		if (X == '\0') break;
		if (Y == '\0') break;
		X = LOWERCASE(X);
		Y = LOWERCASE(Y);
		if (X > Y) return 1;
		if (X < Y) return -1;
	}
    //  If string 1 ended... search the remainder of string 2.
	//  Anything other than a space here means string 1 is smaller.

	if (X == '\0') {
        while(1)  {
			if (Y == '\0') return 0;
			if (Y != ' ') return -1;
		    Y = (unsigned char) *B++;
        }
    }
    
     //  If string 2 ended, search the remainder of string 1.
     // Anything other than a space means string 1 is larger.

	while (1) {
		if(X == '\0') return 0;
		if(X != ' ') return 1;
		X =(unsigned char) *A++;
	}		
}


/**************************************************************************
*                                                                         *
*    Put a string value into an array variable.                           *
*    Is passed the array variable pointer, and the string.                *
*                                                                         *
*    Returns TRUE if successful.                                          *
*                                                                         *
*    If the operation fails, puts an error indicator in the Common block  *
*    and returns FALSE.                                                   *
*                                                                         *
**************************************************************************/
BOOL PutArrayString(xCommon *Common, LPARRVAR ArrayItem, char *TheString) 
{
    ArrayItem->Ainittype = VARTYPE_STRING;  
    ArrayItem->Astrptr = (DWORD) (*lpfnDllServices)(hBatData, 42, TheString, "", 0, 0);
    if ( ArrayItem->Astrptr == 0)
    {
       GlobalFree(Common->Array.lpBuffer);
       Common->Type = 'E';
       Common->ErrorType = -3;
       Common->Long = 104;
       return FALSE;
    } 
    return TRUE;
}

/**************************************************************************
*                                                                         *
*   Fetch the string value from an array item.                            *
*                                                                         *
**************************************************************************/
LPSTR GetArrayString(xCommon *Common, LPARRVAR ArrayItem)
{
  LONG     nResult;

  if (!(lpfnDllServices)) 
     nResult = 0;
  else
  {
     nResult = (*lpfnDllServices)(hBatData, 41, "", "", 0, ArrayItem->Astrptr);
  }
  if (!(nResult))
  {
       GlobalFree(Common->Array.lpBuffer);
       Common->Type = 'E';
       Common->ErrorType = -3;
       Common->Long = 103;
       return NULL;
  }  
  return (LPSTR) (DWORD) nResult;
}

/***************************************************************************
*                                                                          *
*  Get the type code of an array item                                      *
*                                                                          *
***************************************************************************/
long ArrayItemType(LPARRVAR ArrayItem)
{
    long T;
    T = ArrayItem->Ainittype;
    if (T & VARTYPE_STRING) return VARTYPE_STRING;
    if (T & VARTYPE_INT) return VARTYPE_INT;
    if (T & VARTYPE_FLOATNUM) return VARTYPE_FLOATNUM;
    return -1;
}

/***************************************************************************
*                                                                          *
*   ArrayElement(Array, x1, x2, x3, x4, x5)                                *
*                                                                          *
*    Returns a pointer to the specified array element.                     *
*                                                                          *
***************************************************************************/

LPARRVAR ArrayElement(LPRESULTVAR2 arr, WORD x0, WORD x1, WORD x2, WORD x3, WORD x4)
{
  DWORD Offset, Scale, J, K;
  WORD  x[5];

  x[0] = x0;
  x[1] = x1;
  x[2] = x2;
  x[3] = x3;
  x[4] = x4;

  Offset = 0;
  for ( J = 0; J < arr->cDims; J++)
  {
    if (x[J] < 0) return NULL;                 //  subscript out of bounds
    if (x[J] >= arr->wDim[J])  return (NULL); 
    Scale = 1;
    for (K = (J + 1); K < arr->cDims; K++) Scale *= arr->wDim[K];
    Offset += (x[J] * Scale);
  }  
  return ((LPARRVAR) arr->lpBuffer + Offset);
}


/***************************************************************************
*                                                                          *
*    Array Sort Routine                                                    *
*                                                                          *
*    Parameter 2 is the array.  Parameter 3 is the function used for       *
*    string compares.  (CompareString or CompareStringNC)                  *
*                                                                          *
***************************************************************************/
void ArraySorter(xCommon *Common, LPRESULTVAR2 ArrIn, long (*CompString)(char *, char *))
{
    LPARRVAR First;
    LPARRVAR Low;
    LPARRVAR Next;

    long     LowType;
    long     NextType;

    LPSTR    LowString;
    LPSTR    NextString;

    long     Compare;
    char    *Hold;
        
    long     ItemSize;
    long     ItemCount;
    long     ItemCountMinus1;
    
    long     X;
    long     Y;
        
    //------------------------------------------------------------------
    // Compute the distance (in bytes) between the first two items
    //------------------------------------------------------------------
    First  = (LPARRVAR) ArrIn->lpBuffer;
    if ((Next = ArrayElement(ArrIn,1,0,0,0,0)) == NULL) return;
    ItemSize = (Next - First) * sizeof(arrvar);
        
    //-------------------------------------------------------------------
    // Number of items
    //-------------------------------------------------------------------
    ItemCount = ArrIn->wDim[0];
    ItemCountMinus1 = ItemCount-1;
        
    //-------------------------------------------------------------------
    // Get a buffer for temporary storage. big enough to hold one item
    //-------------------------------------------------------------------        
    if (!GetBuffer(Common, ItemSize)) return;
    Hold = Common->Buffer;
        
    //----------------------------------------------------
    // Sort the array.
    //----------------------------------------------------
    for (X=0; X<ItemCountMinus1; X++)
    {
       First = (LPARRVAR) (ArrIn->lpBuffer + (ItemSize * X));
       
       Low = First;
       LowType = ArrayItemType(Low);
       if (LowType == VARTYPE_STRING) 
       {  
          LowString = GetArrayString(Common, Low);
          if (LowString == NULL) return;
       }
       
       Next = Low;           
       for (Y=X+1; Y<ItemCount; Y++)
       {
            Next = (LPARRVAR) (((char *) Next) + ItemSize);
            NextType = ArrayItemType(Next);
            if (NextType == VARTYPE_STRING) 
            { 
                NextString = GetArrayString(Common, Next);
                if (NextString == NULL) return;
            }
            
            Compare = 0;
            switch (LowType)
            {
                case -1:
                {
                     if (NextType != -1) Compare = 1;
                     break;
                }   
                  
                case VARTYPE_INT:
                {
                    if (NextType == VARTYPE_INT)
                    {
                        if (Next->Ax < Low->Ax) Compare = 1;
                        break;
                    }
                    if (NextType == VARTYPE_FLOATNUM)
                    {
                        if (Next->Aflt < (double) Low->Ax ) Compare = 1;
                        break;
                    }
                    break;
                }
    
                case VARTYPE_FLOATNUM:
                {
                    if (NextType == VARTYPE_FLOATNUM)
                    {
                        if (Next->Aflt <  Low->Aflt) Compare = 1;
                        break;
                    }
                    if (NextType == VARTYPE_INT)
                    {
                        if ((double) Next->Ax < Low->Aflt) Compare = 1;
                        break;
                    }
                    break;
                } 

                case VARTYPE_STRING:
                {
                    if (NextType == VARTYPE_STRING)
                    {
                        Compare = (*CompString)(LowString, NextString);
                        break;
                    }
                    if (NextType != -1) Compare = 1;
                    break;
                }      
            }
            if (Compare == 1)
            {
                 Low = Next;
                 LowType = NextType;
                 LowString = NextString;
            }
       }                  
       
       if (Low != First)
       {
            CopyMemory(Hold,Low,ItemSize);
            CopyMemory(Low,First,ItemSize);
            CopyMemory(First,Hold,ItemSize);
       }
    }
    return;
}


/***************************************************************************
*                                                                          *
*     Debug Functions                                                      *
*                                                                          *
*      MBSN ... Display a decimal number.                                  *
*      MBSF ... Display a floating-point number                            *
*      MBSX ... Display a hexidecimal number.                              *
*      MBSS ... Display a string.                                          *
*                                                                          *
***************************************************************************/ 
 void FAR PASCAL MBSN(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"%d",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }

 void FAR PASCAL MBSX(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"x%08lX",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void FAR PASCAL MBSF(char *Title, double Flt)
 {
     char Debug[128];
     long I;
     long J;
     
     I = Flt;
     if (Flt < 0) Flt = 0-Flt;
     J = (Flt * 1000.0);
     wsprintf(Debug,"%d.%03.3d",I,J%1000);
     MessageBox(GetFocus(),Debug,Title,MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void FAR PASCAL MBSS(char *Title, char *Text)
{
     MessageBox(GetFocus(), Text, Title, MB_OK | MB_SETFOREGROUND);
}
